import rumd
import time


def rumd_LJ(params):
    # default parameter values
    particles_per_block = 16
    threads_per_particle = 8
    thermostat = 0                  # 0=none 1=Nose-Hoover
    num_io_blocks = 1L              # Number of (logarthmic) configuration blocks
    io_block_size = 1L
    restartBlock = -1               # which block to restart from (-1 means no restart)
    energyWriteInterval = 200       # Interval (timesteps) between dumping of energies (0 means do not write energies)
    momentum_reset_interval = 100   # 0 to disable

    Ti = 1.                         # Initial temperature
    Tf = 1.                         # Final temperature
    dt = 0.0025
    skin = 0.3
    Rscal = 1.                      # Scaling of initial configuration
    relaxation = 0.3
    
    Tf_assigned = False
    writeTrajectory = True

    # user-provided parameter values, which if present override the above ones
    for key in params.keys():
        cmdStr =  key + " = params['" + key + "']"
        #print cmdStr
        exec(cmdStr)


    if (not Tf_assigned):
        Tf = Ti

    # Create a new sample
    simBox = rumd.RectangularSimulationBox()
    S = rumd.Sample(particles_per_block, threads_per_particle)
    S.SetSimulationBox(simBox)


    # Create an integrator.
    itg = rumd.IntegratorNVT()

    # Setup the integrator parameters
    itg.SetTimeStep(dt)
    itg.SetRelaxationTime(relaxation)
    itg.SetTargetTemperature(Ti)
    itg.SetThermostatOn(thermostat)
    S.SetIntegrator(itg)
    
    # Create LJ pair-potential
    potential = rumd.Pot_LJ_12_6()

    # Setup parameters for Lennard-Jones interactions ( Kob & Andersen )
    # i  j   Sig   Eps  Rcut (in units of Sigma_ij)
    potential.SetParams(0, 0, 1.00, 1.00, 2.5)
    potential.SetParams(0, 1, 0.80, 1.50, 2.5)
    potential.SetParams(1, 0, 0.80, 1.50, 2.5)
    potential.SetParams(1, 1, 0.88, 0.50, 2.5)
    potential.SetNbListSkin(skin)
    S.SetPotential(potential)

    #Read configuration & Copy it to device
    initialTimeStep = 0L

    if(restartBlock < 0):
        S.ReadConf("start.xyz.gz")
        S.IsotropicScaleSystem(Rscal)
        itg.SetMomentumToZero()
    else:
        S.ReadRestartConf(restartBlock)
        initialTimeStep = long( restartBlock*io_block_size )


    potential.WritePotentials(S.GetSimulationBox())
    
    # Setup of various stuff
    S.CreateLogLinManagers(io_block_size, writeTrajectory)
    S.SetEnergyLinearWriteInterval(energyWriteInterval)

    # need to add access to trajectoryWriter metaData
    S.trajectoryWriter.metaData.has_pot_energies = True;
    S.trajectoryWriter.metaData.has_velocities = True;
    S.trajectoryWriter.metaData.has_virials = True;

    if(num_io_blocks > 9999):
        raise "Error: Number of IO blocks cannot be more than 9999"
        

    num_iter = long ( num_io_blocks*io_block_size + 1 )
    dT = (Tf - Ti)/float(num_iter);

    print "Total number of timestep(s) is %d, " % num_iter
    if(restartBlock >= 0):
        print "restart: starting from iteration %lu\n" %initialTimeStep
    print "dumping energies every %d timestep (every %f time-unit).\n" %(energyWriteInterval, energyWriteInterval*itg.GetTimeStep()) 

        
    startTime = time.time()

    for i in xrange( initialTimeStep, num_iter):
        S.CalcF() # Calculate forces
        S.NotifyFileWriters(i) # Called here to ensure synchronization
        itg.Integrate() # Integrate NVE or NVT dynamics.
            
        if ( momentum_reset_interval > 0 and i % momentum_reset_interval==0 ):
            itg.SetMomentumToZero();
    # end main loop

    endTime = time.time()
    gpu_time = endTime - startTime


    print "\nProcessing time GPU (s): %f" % gpu_time
    print " N  thermostat  pb  tp  ms/timestep:"
    t_per_step = 0.
    if (num_iter > 0):
        t_per_step = 1000.0*gpu_time/num_iter
    print "%d %d   %d   %d  %f\n" %( S.GetNumberOfParticles(), thermostat, particles_per_block, threads_per_particle, t_per_step)
  
    # Update forces and velocities and copy everything back from device 
    S.CalcF(potential)  
    S.WriteConf("confout.xyz")
  
if __name__ == "__main__":


    # the following choices correspond to doing test0 with pb=16, tp=1
    userParams = {}
    userParams["num_io_blocks"] = 1
    userParams["io_block_size"] = 100
    userParams["particles_per_block"] = 16
    userParams["threads_per_particle"] = 1
    userParams["Ti"] = 0.500
    userParams["dt"] = 0.005
    userParams["writeTrajectory"] = False
    userParams["thermostat"] = 1
    userParams["energyWriteInterval"] = 100

    # the following changes make it test1 (same pb, tp)
    # comment them out to get test0
    userParams["io_block_size"] = 10000
    userParams["dt"] = 0.002
    userParams["thermostat"] = 0
    userParams["energyWriteInterval"] = 50

    # extra, for comparing performance
    userParams["num_io_blocks"] = 1
    
    rumd_LJ(userParams)
