#ifndef CONFIGURATIONMETADATA_H
#define CONFIGURATIONMETADATA_H

#include <string>
#include <map>
#include <vector>

#include "rumd/rumd_base.h"

#include "rumd/LogLin.h"
#include "rumd/RUMD_Error.h"

// this class handles two kinds of metaData for configuration files
// (1) specific data in the comment line, such as the box-length (this is
// specfically needed for reading, not so much for writing since it is available
// to the sample class anyway)
// (2) which per-particle data is in the file (specified as the "columns" entry 
// in the comment line)

class ConfigurationMetaData
{
  friend class ConfigurationWriterReader;
public:
  ConfigurationMetaData():
    precision(6),
    rumd_conf_ioformat(0), // refers to a file, so could be different from the current rumd_conf_ioformat
    numTypes(1), // refers to a file
    timeStepIndex(0),
    Ps(0.),
    Pv(0.),
    dt(0.),
    massOfType(),
    blockSize(0),
    logLin(),
    fields_present(),
    simulationBoxInfoStr(""),
    integratorInfoStr(""),
    found_integrator(false)
  {
    fields_present["images"] = false;
    fields_present["velocities"] = false;
    fields_present["forces"] = false;
    fields_present["pot_energies"] = false;
    fields_present["virials"] = false;
    fields_present["logLin"] = false;
    for(unsigned int tdx = 0;tdx < MaxNumTypes; ++tdx)
      massOfType[tdx] = 1.;
  }
  // the function which parses the comment line for configuration files
  void ReadMetaData(char *commentLine, bool verbose);
  void SetLogLin(LogLin set_logLin) {this->logLin = set_logLin;} 

  void SetPrecision(int set_precision) { precision = set_precision; }
  void Set(const std::string& key, bool val) {
    if(fields_present.find(key) == fields_present.end())
      throw RUMD_Error("EnergiesMetaData","Set",std::string("Unrecognized key:") + key);
    fields_present[key] = val;
  }
  bool Get(const std::string& key) {
    if(fields_present.find(key) == fields_present.end())
      throw RUMD_Error("EnergiesMetaData","Set",std::string("Unrecognized key:") + key);
    return fields_present[key];
  }
  
  bool IntegratorFound() const { return found_integrator; }    
  std::string GetIntegratorInfoStr() const { return integratorInfoStr; }
  std::string GetSimulationBoxInfoStr() const { return simulationBoxInfoStr; }

  unsigned int GetIO_Format() const { return rumd_conf_ioformat; }
  
  float GetThermostatPs() const { return Ps; }
  unsigned int GetNumTypes() const {return numTypes; }
  unsigned int GetTimeStepIndex() const { return timeStepIndex; }
  float GetDt() const { return dt; }
  const LogLin& GetLogLin() const { return logLin; }
  float GetMassOfType(unsigned typeIdx) const {
    if(typeIdx >= MaxNumTypes)
      throw RUMD_Error("ConfigurationMetaData","GetMassOfType","typeIdx out of range");
    return massOfType[typeIdx];
}

private:
  unsigned int precision;
  // might want to add a separate precision for energies/virials
  unsigned int rumd_conf_ioformat;
  unsigned int numTypes;
  unsigned int timeStepIndex;
  float Ps;
  float Pv;
  float dt;
  float massOfType[MaxNumTypes];
  unsigned int blockSize;
  LogLin logLin;

  std::map<std::string, bool> fields_present;

  // info strings for simulation box and integrator (ioformat 2)
  std::string simulationBoxInfoStr;
  std::string integratorInfoStr;
  bool found_integrator;
};


#endif // CONFIGURATIONMETADATA_H
