#ifndef SAMPLE_H
#define SAMPLE_H

/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include <vector>

#include "rumd/rumd_base.h"
#include "rumd/rumd_technical.h"
#include "rumd/ConfigurationWriterReader.h"
#include "rumd/LogLinOutputManager.h"
#include "rumd/EnergiesOutputManager.h"
#include "rumd/EnergiesMetaData.h"

#include "rumd/SimulationBox.h"
#include "rumd/ParticleData.h"
#include "rumd/NeighborList.h"
#include "rumd/KernelPlan.h"

class Integrator; class Potential;
class MoleculeData;

class Sample{
  friend class ConfigurationWriterReader;

private:
  unsigned int particles_per_block;
  unsigned int threads_per_particle;
  unsigned int degreesOfFreedom;
  unsigned long int blockSize;
  bool allow_tp_280_large;

  std::string trajectoryDir;
  bool writers_initialized;
  std::vector<LogLinOutputManager*> outputManagers;

  LogLinOutputManager* restartLogLin;
  LogLinOutputManager* trajectoryLogLin;
  EnergiesOutputManager* energiesLogLin; 
  std::vector<Potential*> potentialList;

  // Input/Output
  bool verbose;
  float4 CalculateCenterOfMass( float4* position, float4* velocity, int moleculeIndex );

  ConfigurationWriterReader* restartWriter;
  ConfigurationWriterReader* trajectoryWriter;
  void CreateDefaultOutputManagers();

 protected:
  SimulationBox* simulationBox;
  bool own_sim_box;
  Integrator* itg; 
  ParticleData particleData;
  MoleculeData* moleculeData;
  KernelPlan kPlan;
  SortingScheme sortingScheme;
  
 public:
  Sample( unsigned int pb, unsigned int tp, bool allow_tp_280_large );
  Sample(const Sample& S);
  Sample* Copy();
  ~Sample();
  Sample& operator=(const Sample& S);

  void InitializeOutputManagers(unsigned long int timeStepIndex);
  void NotifyOutputManagers(unsigned long int timeStepIndex);
  void TerminateOutputManagers(); // close open files etc
  void ReadConf( const std::string& filename, bool init_itg=true );
  void ReadRestartConf(int restartBlock, unsigned int numDigitsBlkIdx=4 );
  void WriteConf(const std::string& filename, const std::string& mode="w"){ restartWriter->Write(this, filename, mode); }
  void AddOutputManager(LogLinOutputManager* om);
 
  float CalcF(bool calc_stresses = false, bool report_time = false);
  void IsotropicScaleSystem( float Rscal );
  void AnisotropicScaleSystem( float Rscal, int dir );
  void CopySimulationDataToDevice() const;
  void CopySimulationDataFromDevice() const;
  void AddPotential(Potential* potential);
  void SubtractDOFs( unsigned i ){ degreesOfFreedom -= i; };
  void SetMoleculeData(MoleculeData* md) { moleculeData = md; }
  // Spatial sorting
  void SortParticles();
  
  // Set methods
  void SetOutputBlockSize(unsigned long int blockSize);
  void SetWriteEnergies(bool writeEnergies);
  void SetWriteTrajectory(bool writeTrajectory);
  void SetWriteRestart(bool writeRestart);
  void SetTrajectoryLinearWriteInterval(unsigned long int writeInterval);
  void SetTrajectoryLogLinParams(unsigned int base, unsigned int maxInterval);
  void SetEnergyLinearWriteInterval(unsigned int writeInterval); 
  void SetEnergyLogLinParams(unsigned int base, unsigned int maxInterval); 
  void SetNumberOfParticles(unsigned int num_part);
  void SetSimulationBox(SimulationBox* simBox, bool own_this_box=false);
  void SetIntegrator(Integrator* newItg);
  void SetPotential(Potential* potential);
  void SetVerbose(bool vb);
  void SetOutputDirectory(const std::string& trajectoryDir);
  void SetNumberOfDOFs( unsigned i ){ degreesOfFreedom = i; };

  // the following is temporary
  void SetConfIO_Format(unsigned int ioformat) {
    trajectoryWriter->SetIO_Format(ioformat);
    restartWriter->SetIO_Format(ioformat);
  }
  void SetPB_TP( unsigned int pb, unsigned int tp );
  void SetSortingScheme(SortingScheme ss) {sortingScheme = ss;}

  // Get methods.
  SimulationBox* GetSimulationBox() const { return simulationBox; }  
  ParticleData* GetParticleData() { return &particleData; }
  const ParticleData* GetParticleDataConst() const { return &particleData; }
  MoleculeData* GetMoleculeData() { return moleculeData; }
  Integrator* GetIntegrator() const { return itg; }
  std::vector<Potential*>* GetPotentials() { return &potentialList; }  
  const KernelPlan& GetKernelPlan() const { return kPlan; }
  const std::string& GetOutputDirectory() const { return trajectoryDir; }

  unsigned int GetParticlesPerBlock() const { return particles_per_block; }
  unsigned int GetThreadsPerParticle() const { return threads_per_particle; }
  unsigned int GetNumberOfDOFs() const { return degreesOfFreedom; }
  unsigned int GetNumberOfParticles() const { return particleData.GetNumberOfParticles(); }
  unsigned int GetNumberOfVirtualParticles() const { return particleData.GetNumberOfVirtualParticles(); }
  unsigned int GetNumberOfTypes() const { return particleData.GetNumberOfTypes(); }
  unsigned int GetNumberThisType(unsigned int type) const { return particleData.GetNumberThisType(type); }
  unsigned long int GetOutputBlockSize() const { return blockSize; }

  float GetMass(unsigned int type) const { return particleData.GetMass(type); }
  float GetMeanMass() const { float meanMass = 0; for(unsigned i=0; i < GetNumberOfTypes(); i++){ meanMass += GetMass(i); } meanMass /= GetNumberOfTypes(); return meanMass; }

  ConfigurationMetaData& GetTrajectoryConfMetaData() {return trajectoryWriter->metaData; }
  EnergiesOutputManager& GetEnergiesOutputManager() {return (*energiesLogLin);}

  double GetPotentialEnergy(bool copy=true) const;  
  double GetVirial(bool copy=true) const;
  SortingScheme GetSortingScheme() const { return sortingScheme; }
};


#endif // SAMPLE_H
