/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/
#include "rumd/Device.h"

#include <iostream>
#include <cstdlib>
#include <vector>
#include <sys/utsname.h>
#include <sstream>

#include "rumd/RUMD_Error.h"


Device::~Device() {
  cudaError err = cudaThreadExit();
  if (err != cudaSuccess )
    throw RUMD_Error("Device","~Device",std::string("[cudaThreadExit] ") + cudaGetErrorString(err));
  std::cout << "CUDA: gpu released" << std::endl; // TODO: remove
}

void Device::Report()
{
  int dev;
  cudaGetDevice(&dev);
  cudaDeviceProp deviceProp;
  cudaError err = cudaGetDeviceProperties(&deviceProp, dev);
  if (err != cudaSuccess)
    throw RUMD_Error("Device","Report",std::string("[cudaGetDeviceProperties] ")+cudaGetErrorString(err));

  struct utsname buf;
  uname(&buf);
  std::cout
    << "CUDA:"
    << " node=" << buf.nodename
    << " device=" << dev
//    << " pciBusID=" << deviceProp.pciBusID
//    << " pciDeviceID=" << deviceProp.pciDeviceID
    << " name=\"" << deviceProp.name << "\""
    << std::endl;
}

const std::string Device::GetDeviceName() {
  int dev;
  cudaGetDevice(&dev);
  cudaDeviceProp deviceProp;
  cudaError err = cudaGetDeviceProperties(&deviceProp, dev);
  if (err != cudaSuccess)
    throw RUMD_Error("Device","Report",std::string("[cudaGetDeviceProperties] ")+cudaGetErrorString(err));
  std::ostringstream outputStr;
  outputStr << deviceProp.name;
  return outputStr.str();
}

Device& Device::GetDevice() {
  static Device instance;
  return instance;
}

void Device::Init(int argc, char *argv[])
{
  int deviceCount;
  cudaError err = cudaGetDeviceCount(&deviceCount);
  if (err != cudaSuccess)
    throw RUMD_Error("Device","Init",std::string("[cudaGetDeviceCount]") + cudaGetErrorString(err));

  if (deviceCount == 0)
    throw RUMD_Error("Device","Init","error: no devices supporting CUDA.");

  if (deviceCount == 1) {
    // just one device or an emulated device present, no choice
    int dev=0;
    cudaDeviceProp deviceProp;
    cudaError err = cudaGetDeviceProperties(&deviceProp, dev);
    if (err != cudaSuccess)
      throw RUMD_Error("Device","Init",std::string("[cudaGetDeviceProperties]") + cudaGetErrorString(err));
    if (deviceProp.major < 1) 
      throw RUMD_Error("Device","Init","error: device does not support CUDA.");
    if ((deviceProp.major == 1) && (deviceProp.minor < 3))
      throw RUMD_Error("Device","Init","error: device compute capability >= 1.3 needed");
    cudaSetDevice(dev);
  }
  else {
    // several devices present, so make list of usable devices
    // and have one choosen among the currently available ones
    std::vector<int> usable_devices;
    for (int dev=0; dev<deviceCount; dev++) { 
      cudaDeviceProp deviceProp;
      cudaError err = cudaGetDeviceProperties(&deviceProp, dev);
      if (err != cudaSuccess)
	throw RUMD_Error("Device","Init",std::string("[cudaGetDeviceProperties] ") + cudaGetErrorString(err));
      if (((deviceProp.major > 1) ||
	 ((deviceProp.major == 1) && (deviceProp.minor >= 3))) && 
	(deviceProp.multiProcessorCount >= 2) &&
	 (deviceProp.computeMode != cudaComputeModeProhibited)) {
        usable_devices.push_back(dev); 
      }
    }
    if (usable_devices.size() == 0)
      throw RUMD_Error("Device","Init","error: no usable devices supporting CUDA.");
    cudaError err = cudaSetValidDevices(&usable_devices[0], usable_devices.size());
    if (err != cudaSuccess )
      throw RUMD_Error("Device","Init",std::string("[cudaSetValidDevices] ") + cudaGetErrorString(err));
    // trigger device initialization by a non-device management function call
    cudaError err2 = cudaThreadSynchronize();
    if (err2 != cudaSuccess )
      throw RUMD_Error("Device","Init",std::string("[cudaThreadSynchronize] ") + cudaGetErrorString(err));
    Report();
  }
}
