/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include <iostream>

#include "rumd/rumd_utils.h"
#include "rumd/PairPotential.h"
#include "rumd/SimulationBox.h"
#include "rumd/NeighborList.h"
#include <cmath>

using namespace std;

///////////////////////////////////////////
// CONSTRUCTION/DESTRUCTION 
///////////////////////////////////////////

NeighborList::NeighborList() : simBox(0) {
  skin = 0.3f; // default value
  max_cutoff = 0.f;
  elapsedTimeLastUpdate = 0.f;
  
  numBlocks = 0;
  numBlocksParticles = 0;
  numVirtualParticles = 0;
  maxNumNbrBlocks = 0;
  // it's a good idea to initialize array pointers to zero for debugging ...
  h_ExclusionType = 0;
  h_ExclusionList = 0;
  h_num_nbr_blocks = 0; 
  NbList = 0;
  d_savedNeighborList = 0;
  cell_list = 0;
  d_last_r = 0;
  d_Rc = 0;
  d_dR = 0;
  d_ExclusionType = 0;
  d_ExclusionList = 0;
  d_OldExclusionList = 0;
  d_num_nbr_blocks = 0;

  // for keeping track of allocation. Number in the comment indicates how many
  // arrays share this number (allocated together)
  numberAllocatedUnits["NbList"] = 0; // 1
  numberAllocatedUnits["d_savedNeighborList"] = 0; // 1

  numberAllocatedUnits["d_last_r"] = 0; // 1
  numberAllocatedUnits["cell_list"] = 0; // 1
  numberAllocatedUnits["d_Rc_d_dR"] = 0; // 2 (d_Rc and d_dR)
  numberAllocatedUnits["num_nbr_blocks"] = 0; // 2 (h_ and d_)

  numberAllocatedUnits["ExclusionType"] = 0; // 2 (h_ and d_)
  numberAllocatedUnits["ExclusionList"] = 0; // 3 (h_ and d_ and d_Old)

  // Here we allocate memory that doesn't depend on the system size
  if ( cudaMalloc( (void**) &d_shell, sizeof(float) * MaxNumTypes * MaxNumTypes ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("NeighborList","NeighborList","Malloc failed on d_shell") );
  
  if ( cudaMalloc( (void**) &d_lastBoxShift, sizeof(float) ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("NeighborList","NeighborList","Malloc failed on d_lastBoxShift") );
  
  float h_lastBoxShift = 0.f;
  cudaMemcpy(d_lastBoxShift, &h_lastBoxShift, sizeof(float), cudaMemcpyHostToDevice);
  
  if ( cudaMalloc( (void**) &d_dStrain, sizeof(float2) ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("NeighborList","NeighborList","Malloc failed on d_dStrain") );
  
  float2 h_dStrain = {0.f,0.f};
  cudaMemcpy(d_dStrain, &h_dStrain, sizeof(float2), cudaMemcpyHostToDevice);
  
  if( cudaMalloc( (void**) &d_updateNbList, sizeof(int) ) == cudaErrorMemoryAllocation )
    throw( RUMD_Error("NeighborList","NeighborList","Malloc failed on d_updateNbList") );


}

NeighborList::~NeighborList(){

  if(numberAllocatedUnits["NbList"] > 0)
    cudaFree(NbList);

  if(numberAllocatedUnits["d_savedNeighborList"] > 0)
    cudaFree(d_savedNeighborList);

  if(numberAllocatedUnits["d_last_r"] > 0)
    cudaFree(d_last_r);

  if(numberAllocatedUnits["cell_list"] > 0)
    cudaFree(cell_list);

  if(numberAllocatedUnits["d_Rc_d_dR"] > 0) {
    cudaFree(d_Rc);
    cudaFree(d_dR);
  }

  if(numberAllocatedUnits["num_nbr_blocks"] > 0) {
    cudaFreeHost(h_num_nbr_blocks);
    cudaFree(d_num_nbr_blocks);
  }

  if(numberAllocatedUnits["ExclusionType"] > 0) {
    cudaFreeHost(h_ExclusionType);
    cudaFree(d_ExclusionType);
  }
  
  if(numberAllocatedUnits["ExclusionList"] > 0) {
    cudaFreeHost(h_ExclusionList);
    cudaFree(d_ExclusionList);
    cudaFree(d_OldExclusionList);
  }

  cudaFree(d_shell);
  cudaFree(d_lastBoxShift);
  cudaFree(d_dStrain);
  cudaFree(d_updateNbList);
}

///////////////////////////////////////////
// MEMORY ALLOCATION/DEALLOCATION 
///////////////////////////////////////////

void NeighborList::Initialize(ParticleData* pData, KernelPlan& kPlan, SimulationBox* simBox){
  particleData = pData;
  numParticles = particleData->GetNumberOfParticles();
  kp = kPlan;
  this->simBox = simBox;
  
  numBlocks = kp.num_blocks;
  numVirtualParticles = kp.num_virt_part;
  numBlocksParticles = numBlocks * numVirtualParticles;
  particlesPerBlock = kp.threads.x;
  
  /*unsigned int estMaxNumNbrBlocks = EstimateMaxNumNbrBlocks(ss);
  if(estMaxNumNbrBlocks > numBlocks/2)
    estMaxNumNbrBlocks = numBlocks;
    SetMaxNumNbrBlocks(estMaxNumNbrBlocks);*/

  SetMaxNumNbrBlocks(numBlocks); // Allocates neighbor-list itself
  
  AllocateLastPositions();
  
  AllocateNumBlocks(); // arrays with length numBlocks
  
  AllocateExclusionType( particleData->GetNumberOfTypes() );

  // we don't allocate exclusion list right away, only when needed
  // but if it *has* previously been allocated then it may be necessary to
  // re-allocate it
  if(numberAllocatedUnits["ExclusionList"] > 0)
    AllocateExclusionList();
  
  
  ResetNeighborList(); // Put into Allocate functions as well?
}


unsigned int NeighborList::EstimateMaxNumNbrBlocks(SortingScheme ss) {
  
  // determine maxNumNbrBlocks according to sorting scheme
  float Lx = simBox->GetLength(0), Ly = simBox->GetLength(1), Lz = simBox->GetLength(2);
  float rc_s = max_cutoff + skin;
    
  // SORT_X
  unsigned int numNbrBlocksX = 1 + 2*ceil(1.5*numParticles*rc_s/particlesPerBlock/Lx);
  // factor 1.5 is to cover de-sorting
  std::cout << "maxNumNbrBlocks (SORT_X): " << numNbrBlocksX << std::endl;

  // SORT_Y
  float areaPerBlock = Lx*Ly*particlesPerBlock/numParticles;
  float d = sqrt(areaPerBlock);
  // radius is rc + half a block diagonal a couple of block widths to account 
  // for the ragged perimeter, round up to 3 d
  float radius = rc_s + 3.*d;
  unsigned int numNbrBlocksXY = ceil(1.3*M_PI*radius*radius/areaPerBlock);
  std::cout << "maxNumNbrBlocks (SORT_XY): " << numNbrBlocksXY << std::endl;
  // factor 1.3 for de-sorting

  // SORT_XYZ
  float volumePerBlock = Lx*Ly*Lz*particlesPerBlock/numParticles;
  d = pow(volumePerBlock, 1./3);
  // rc+half-diagonal+one block width, round to 2
  radius  = rc_s + 2.*d;
  unsigned int numNbrBlocksXYZ = ceil(1.3*4.*M_PI/3.*radius*radius*radius / volumePerBlock);
  // factor 1.3 for de-sorting
  std::cout << "numNbrBlocks (SORT_XYZ): " << numNbrBlocksXYZ << std::endl;

  unsigned int estMaxNumNbrBlocks;
  
  switch(ss) {
    case SORT_X:
      estMaxNumNbrBlocks = numNbrBlocksX;
      break;
    case SORT_XY:
      estMaxNumNbrBlocks = numNbrBlocksXY;
      break;
    case SORT_XYZ:
      estMaxNumNbrBlocks = numNbrBlocksXYZ;
      break;
  default:
    estMaxNumNbrBlocks = numNbrBlocksXYZ;
  }
  return estMaxNumNbrBlocks;
}

void NeighborList::SetMaxNumNbrBlocks(unsigned max_num_nbr_blocks) {
  maxNumNbrBlocks = max_num_nbr_blocks;
  if(maxNumNbrBlocks > numBlocks)
    maxNumNbrBlocks = numBlocks;

  //std::cout << "Set maxNumNbrBlocks=" << maxNumNbrBlocks << std::endl;
  AllocateNeighborList();
  // Re-allocate d_savedNeighborList if it has been allocated
  if(numberAllocatedUnits["d_savedNeighborList"] > 0)
    AllocateSavedNeighborList();

}

void NeighborList::AllocateLastPositions() {

  if(numberAllocatedUnits["d_last_r"] != numVirtualParticles) {

    if( numberAllocatedUnits["d_last_r"] > 0 )
      cudaFree(d_last_r);

    if( cudaMalloc( (void**) &d_last_r, numVirtualParticles * sizeof(float4) ) == cudaErrorMemoryAllocation )
      throw( RUMD_Error("NeighborList","Initialize","Malloc failed on d_last_r") );
    // initialize to zeros
    cudaMemset( d_last_r, 0, numVirtualParticles * sizeof(float4) );
  
    numberAllocatedUnits["d_last_r"] = numVirtualParticles;
  }
}


void NeighborList::AllocateNumBlocks() {

  if(numberAllocatedUnits["num_nbr_blocks"] != numBlocks) {
    if(numberAllocatedUnits["num_nbr_blocks"] > 0) {
      cudaFree(d_num_nbr_blocks);
      cudaFreeHost(h_num_nbr_blocks);
    }

    if( cudaMalloc( (void**) &d_num_nbr_blocks, numBlocks*sizeof(unsigned int) ) == cudaErrorMemoryAllocation )
      throw( RUMD_Error("NeighborList","Initialize","Malloc failed on d_num_nbr_blocks") );
    cudaMemset(d_num_nbr_blocks, 0, numBlocks*sizeof(int));
    if( cudaMallocHost( (void**) &h_num_nbr_blocks, numBlocks*sizeof(unsigned int) ) == cudaErrorMemoryAllocation )
      throw( RUMD_Error("NeighborList","Initialize","Malloc failed on h_num_nbr_blocks") );
    numberAllocatedUnits["num_nbr_blocks"] = numBlocks;
  }
  
  // Bounding box for blocks.
  if(numberAllocatedUnits["d_Rc_d_dR"] != numBlocks) {

    if(numberAllocatedUnits["d_Rc_d_dR"] > 0) {
      cudaFree(d_Rc);
      cudaFree(d_dR);
    }
    
    if( cudaMalloc( (void**) &d_Rc, numBlocks * sizeof(float4) ) == cudaErrorMemoryAllocation )  // Center
    throw( RUMD_Error("NeighborList","Initialize","Malloc failed on d_Rc") );
    
    if( cudaMalloc( (void**) &d_dR, numBlocks * sizeof(float4) ) == cudaErrorMemoryAllocation )  // Distance from surface to center
    throw( RUMD_Error("NeighborList","Initialize","Malloc failed on d_dR") );
    
    numberAllocatedUnits["d_Rc_d_dR"] = numBlocks;
  }
}

void NeighborList::AllocateNeighborList() {
  // these arrays depend on the assumed number of "neighbor blocks", and
  // may need to be reallocated
  if(numberAllocatedUnits["NbList"] != numVirtualParticles * maxNumNbrBlocks) {

    if(numberAllocatedUnits["NbList"] > 0)
      cudaFree(NbList);

    if( cudaMalloc( (void**) &NbList, sizeof(NBTYPEv) * numVirtualParticles * maxNumNbrBlocks ) == cudaErrorMemoryAllocation )
      throw( RUMD_Error("NeighborList","Initialize","Malloc failed on NbList") );
    numberAllocatedUnits["NbList"]  = numVirtualParticles * maxNumNbrBlocks;
  }
  
  if(numberAllocatedUnits["cell_list"] != numBlocks * maxNumNbrBlocks) {
    if(numberAllocatedUnits["cell_list"] > 0)
      cudaFree(cell_list);

    if( cudaMalloc( (void**) &cell_list, sizeof(int) * numBlocks * maxNumNbrBlocks ) == cudaErrorMemoryAllocation )
      throw( RUMD_Error("NeighborList","Initialize","Malloc failed on cell_list") );
    numberAllocatedUnits["cell_list"] = numBlocks * maxNumNbrBlocks;
  }
}


void NeighborList::AllocateSavedNeighborList() {
  if( numberAllocatedUnits["d_savedNeighborList"]  != numVirtualParticles * maxNumNbrBlocks) {
    if (numberAllocatedUnits["d_savedNeighborList"] > 0)
      cudaFree(d_savedNeighborList);
    
    if( cudaMalloc( (void**) &d_savedNeighborList, sizeof(NBTYPEv) * numVirtualParticles * maxNumNbrBlocks ) == cudaErrorMemoryAllocation )
      throw( RUMD_Error("NeighborList","Initialize","Malloc failed on d_savedNeighborList") );
    
    numberAllocatedUnits["d_savedNeighborList"]  = numVirtualParticles * maxNumNbrBlocks;
  }
}


void NeighborList::AllocateExclusionType( unsigned int numberOfTypes ) {
  if( numberAllocatedUnits["ExclusionType"] != numberOfTypes*numberOfTypes) {

    if( numberAllocatedUnits["ExclusionType"] > 0) {
      cudaFreeHost(h_ExclusionType);
      cudaFree(d_ExclusionType);
    }

    if( cudaMallocHost( (void**) &h_ExclusionType, sizeof(unsigned) * numberOfTypes * numberOfTypes ) == cudaErrorMemoryAllocation )
      throw( RUMD_Error("NeighborList","Initialize","Malloc failed on h_ExclusionType") );
    if( cudaMalloc( (void**) &d_ExclusionType, sizeof(unsigned) * numberOfTypes * numberOfTypes ) == cudaErrorMemoryAllocation )
      throw( RUMD_Error("NeighborList","Initialize","Malloc failed on d_ExclusionType") );
    
    numberAllocatedUnits["ExclusionType"] = numberOfTypes*numberOfTypes;

    memset( h_ExclusionType, 0, numberAllocatedUnits["ExclusionType"] * sizeof(unsigned) );
    CopyExclusionTypeToDevice();
  }
}

void NeighborList::AllocateExclusionList() {
  unsigned int nvp_nnb = numVirtualParticles * numBlocks;
  if(numberAllocatedUnits["ExclusionList"] != nvp_nnb) {
    
    if(numberAllocatedUnits["ExclusionList"] > 0) {
      cudaFreeHost(h_ExclusionList);
      cudaFree(d_ExclusionList);
      cudaFree(d_OldExclusionList);
    }
    
    if( cudaMallocHost( (void**) &h_ExclusionList, sizeof(NBTYPEv) * nvp_nnb ) == cudaErrorMemoryAllocation )
      throw( RUMD_Error("NeighborList","Initialize","Malloc failed on h_ExclusionList") );
  
    if( cudaMalloc( (void**) &d_ExclusionList, sizeof(NBTYPEv) * nvp_nnb ) == cudaErrorMemoryAllocation )
      throw( RUMD_Error("NeighborList","Initialize","Malloc failed on d_ExclusionList") );
    
    if( cudaMalloc( (void**) &d_OldExclusionList, sizeof(NBTYPEv) * nvp_nnb ) == cudaErrorMemoryAllocation )
      throw( RUMD_Error("NeighborList","Initialize","Malloc failed on d_OldExclusionList") );
    
    numberAllocatedUnits["ExclusionList"] = nvp_nnb;
    // Clear and copy to GPU
    memset( h_ExclusionList, 0, nvp_nnb * sizeof(NBTYPEv) );
    CopyExclusionListToDevice(); 
 }
}

unsigned int NeighborList::GetMaxUsedNbrBlocks() {
  cudaMemcpy( h_num_nbr_blocks, d_num_nbr_blocks, numBlocks * sizeof(unsigned int), cudaMemcpyDeviceToHost );
  unsigned int max_used_nbr_blocks = h_num_nbr_blocks[0];
  for (unsigned int b = 1; b < numBlocks; b++) {
    if(h_num_nbr_blocks[b] > max_used_nbr_blocks)
      max_used_nbr_blocks = h_num_nbr_blocks[b];
  }
  return max_used_nbr_blocks;
}

/*
void NeighborList::CheckAndReallocate() {
 if(maxNumNbrBlocks < numBlocks) {
    unsigned int max_used_nbr_blocks = GetMaxUsedNbrBlocks();
    while(max_used_nbr_blocks >= maxNumNbrBlocks) {
      std::cout << "Re-allocating neighbor-list " << ((unsigned int) (1.5*maxNumNbrBlocks)) << std::endl;
      SetMaxNumNbrBlocks((unsigned int) (1.5*maxNumNbrBlocks) );    
      UpdateNBlist(false);
      max_used_nbr_blocks = GetMaxUsedNbrBlocks();
    }
 }

}
*/

void NeighborList::CheckAllocatedNbrBlocks() {
  if(maxNumNbrBlocks < numBlocks && GetMaxUsedNbrBlocks() > maxNumNbrBlocks)
    throw RUMD_Error("NeighborList","CheckAllocatedNbrBlocks","Number of neighbor blocks needed exceeds maxNumNbrBlocks");
}

///////////////////////////////////////////
// Main methods.
///////////////////////////////////////////

void NeighborList::UpdateNBlist(bool record_time_taken){
  cudaEvent_t start;
  cudaEvent_t stop;
  elapsedTimeLastUpdate = 0.;
  
  if(record_time_taken) {
    cudaEventCreate( &start );
    cudaEventCreate( &stop );
    cudaEventRecord( start, 0);
  }

  float4* d_r = particleData->d_r;
  float4* d_im = particleData->d_im;
  dim3 threads1(kp.threads.x, 1);
  
  int BlocksPerHyperBlock = 16;
  dim3 HyperGrid((numBlocks+BlocksPerHyperBlock-1)/BlocksPerHyperBlock, 1, 1);
  dim3 HyperThreads(BlocksPerHyperBlock, 1, 1);
  int HyperSharedSize = sizeof(float4)*BlocksPerHyperBlock*2;
  
  RectangularSimulationBox* testRSB = dynamic_cast<RectangularSimulationBox*>(simBox);
  LeesEdwardsSimulationBox* testLESB = dynamic_cast<LeesEdwardsSimulationBox*>(simBox);

  
  // check whether a re-build is necessary
  if(testLESB) {
    update_deltaStrain<<<1, 1>>>(d_lastBoxShift, d_dStrain, testLESB->GetDevicePointer(), skin,  max_cutoff);
    displacement_since_nb_build<<< kp.grid, threads1 >>>( numParticles, d_r, d_last_r, d_im, testLESB, testLESB->GetDevicePointer(), 0.25*skin*skin, d_updateNbList, d_dStrain );
    calc_boundingbox<<< kp.grid, threads1,  kp.shared_size >>>( numParticles, d_r, d_Rc, d_dR, testLESB, testLESB->GetDevicePointer(), max_cutoff + skin, d_updateNbList);
    calc_cell_list<<< HyperGrid, HyperThreads, HyperSharedSize >>>( numBlocks, maxNumNbrBlocks, d_Rc, d_dR, testLESB, testLESB->GetDevicePointer(),  cell_list, d_updateNbList, d_num_nbr_blocks);
    md_update_nblist<<< kp.grid, kp.threads, kp.shared_size >>>( numParticles, maxNumNbrBlocks, d_r, d_im, d_shell, testLESB, testLESB->GetDevicePointer(), 
								 NbList, d_ExclusionList, d_ExclusionType, cell_list, d_updateNbList, d_last_r, d_lastBoxShift, particleData->GetNumberOfTypes() );
  }
  else if(testRSB) {
    displacement_since_nb_build<<< kp.grid, threads1 >>>( numParticles, d_r, d_last_r, d_im, testRSB, testRSB->GetDevicePointer(), 0.25*skin*skin, d_updateNbList, 0 );
    calc_boundingbox<<< kp.grid, threads1,  kp.shared_size >>>( numParticles, d_r, d_Rc, d_dR, testRSB, testRSB->GetDevicePointer(), max_cutoff + skin, d_updateNbList );
    calc_cell_list<<< HyperGrid, HyperThreads, HyperSharedSize >>>( numBlocks, maxNumNbrBlocks, d_Rc, d_dR, testRSB, testRSB->GetDevicePointer(), cell_list, d_updateNbList, d_num_nbr_blocks);
    md_update_nblist<<< kp.grid, kp.threads, kp.shared_size >>>( numParticles, maxNumNbrBlocks, d_r, d_im, d_shell, testRSB, testRSB->GetDevicePointer(), 
								 NbList, d_ExclusionList, d_ExclusionType, cell_list, d_updateNbList, d_last_r, 0, particleData->GetNumberOfTypes() );
  }
  else 
    throw RUMD_Error("NeighborList", "UpdateNBlist", "Unrecognized Simulation box type");

  
  CheckAllocatedNbrBlocks(); // if maxNumNbrBlocks==numBlocks then does nothing

  if(record_time_taken) {
    cudaEventRecord( stop, 0 );
    cudaEventSynchronize( stop );
    cudaEventElapsedTime( &elapsedTimeLastUpdate, start, stop);
    cudaEventDestroy(start);
    cudaEventDestroy(stop);
  }
}

void NeighborList::UpdateExclusionListAfterSorting( unsigned int* new_index ){
  if(numberAllocatedUnits["ExclusionList"] > 0) {
    cudaMemcpy( d_OldExclusionList, d_ExclusionList, sizeof(NBTYPEv) * numBlocks * numVirtualParticles, cudaMemcpyDeviceToDevice );
    cudaMemset( d_ExclusionList, 0, sizeof(NBTYPEv) * numBlocks * numVirtualParticles );
    UpdateMyExclusionListAfterSorting<<< kp.grid, kp.threads.x >>>( new_index, d_OldExclusionList, d_ExclusionList, numParticles );
  }
}

void NeighborList::ResetNeighborList(){
  cudaMemcpy(d_updateNbList, &(kp.grid.x), sizeof(unsigned int), cudaMemcpyHostToDevice);
}

void NeighborList::SaveNeighborList(){
  AllocateSavedNeighborList(); // only allocates if not already done
  cudaMemcpy( d_savedNeighborList, NbList, numBlocksParticles * sizeof(NBTYPEv), cudaMemcpyDeviceToDevice );
}

void NeighborList::RestoreNeighborList(){
  if(numberAllocatedUnits["d_savedNeighborList"] != numberAllocatedUnits["NbList"])
    throw RUMD_Error("NeighborList","RestoreNeighborList","Improperly allocated d_savedNeighborList; perhaps SaveNeighborList has not been called");
  cudaMemcpy( NbList, d_savedNeighborList, numBlocksParticles * sizeof(NBTYPEv), cudaMemcpyDeviceToDevice );
}

///////////////////////////////////////////
// Set methods
///////////////////////////////////////////

void NeighborList::SetCutoffs(PairPotential* potential){
  max_cutoff = 0.f;
  for (unsigned int i=0; i<MaxNumTypes; i++){
    for (unsigned int j=0; j<MaxNumTypes; j++){ 
      unsigned int k = i * MaxNumTypes + j;
      double Rcut = (double) potential->GetPotentialParameter(i,j,0);
      h_shell[k] = Rcut + skin;
      if(Rcut > max_cutoff) max_cutoff = Rcut;
    }
  }  
  
  cudaMemcpy( d_shell, h_shell, sizeof(float)*MaxNumTypes*MaxNumTypes, cudaMemcpyHostToDevice);
  // may need to change maxNumNbrBlocks if less than numBlocks
  ResetNeighborList();
}

void NeighborList::SetCutoffs(PairPotential* potential, float set_skin){ 
  skin = set_skin; SetCutoffs(potential); 
}

void NeighborList::SetExclusionList( NBTYPEv* newList ){
  AllocateExclusionList(); // (if not already done)
  for( unsigned int i=0; i < numBlocksParticles; i++ )
    h_ExclusionList[i] = newList[i];
  
  cudaMemcpy( d_ExclusionList, h_ExclusionList, sizeof(NBTYPEv) * numBlocksParticles, cudaMemcpyHostToDevice );
}
 
// Excludes two particles.
void NeighborList::SetExclusion( unsigned particleI, unsigned particleJ ){
  AllocateExclusionList(); // if not already done

  unsigned a = particleI / particlesPerBlock;
  unsigned b = particleJ / particlesPerBlock;
  
  unsigned ofnblist_i = a * numBlocks * particlesPerBlock;
  unsigned ofnblist_j = b * numBlocks * particlesPerBlock;
      
  unsigned ofnblist = ofnblist_i + b * particlesPerBlock + ( particleI % particlesPerBlock );
  h_ExclusionList[ofnblist].x |= ( 1 << ( particleJ % particlesPerBlock ) );
  
  ofnblist = ofnblist_j + a * particlesPerBlock + ( particleJ % particlesPerBlock );
  h_ExclusionList[ofnblist].x |= ( 1 << ( particleI % particlesPerBlock ) );      
}

// Exclude two given particle pair interactions based on type.
void NeighborList::SetExclusionType( unsigned type0, unsigned type1 ){
  if(numberAllocatedUnits["ExclusionType"] == 0)
    throw RUMD_Error("NeighborList","SetExclusionType","Object is not yet initialized, probably because the potential object has not been associated with a sample object yet.");
  unsigned numberOfTypes = particleData->GetNumberOfTypes();
  
  h_ExclusionType[type0*numberOfTypes+type1] = 1;
  // For now symmetric.
  h_ExclusionType[type1*numberOfTypes+type0] = 1;
  
  CopyExclusionTypeToDevice();
}

// Exclude two given particle pair interactions based on bond and/or constraint.
void NeighborList::SetExclusionBond(uint1* h_btlist, uint2* h_blist, unsigned num_part, unsigned num_virt_part, unsigned num_bonds, unsigned etype){
  if( num_bonds == 0 )
    throw( RUMD_Error("NeighborList","SetExclusionBond","No bonds have been read") );

  AllocateExclusionList(); // if not already done
    

  for( unsigned i = 0; i < num_bonds; i++){
    unsigned userBondType = h_btlist[i].x;
    
    if( userBondType == etype )
      SetExclusion(h_blist[i].x, h_blist[i].y);
    
  }
  CopyExclusionListToDevice();
}

// Exclude two given particle pair interactions based on the angle.
void NeighborList::SetExclusionAngle(uint4 *h_alist, unsigned num_angles){
  if( num_angles == 0 )
    throw( RUMD_Error("NeighborList","SetExclusionAngle","No angles have been read") );
  
  AllocateExclusionList(); // if not already done

  // The last two particles in an angle.
  for ( unsigned i=0; i < num_angles; i++ )
    SetExclusion(h_alist[i].x, h_alist[i].z);
  
  CopyExclusionListToDevice();
}

// Exclude two given particle pair interactions based on dihedral.
void NeighborList::SetExclusionDihedral(uint4* h_dlist, unsigned num_dihedrals){
  if( num_dihedrals == 0 )
    throw( RUMD_Error("NeighborList","SetExclusionDihedral","No dihedrals have been read") );
  
  AllocateExclusionList(); // if not already done

  // The last two particles in a dihedral.    
  for ( unsigned i=0; i < num_dihedrals; i++ )
    SetExclusion(h_dlist[i].x, h_dlist[i].w);
  
  CopyExclusionListToDevice();
}

// Exclude all pair interactions in the same molecule.
void NeighborList::SetExclusionMolecule(int1* h_mlist, unsigned molindex, unsigned max_num_uau, unsigned num_mol){
  AllocateExclusionList(); // if not already done

  unsigned i = molindex*max_num_uau + num_mol;

  for ( unsigned n=i; n<i+max_num_uau-1; n++ ){
    int ai = h_mlist[n].x;

    if ( ai == - 1) break;

    for ( unsigned m=n+1; m<i+max_num_uau; m++ ){

      int bi = h_mlist[m].x;

      if ( bi == -1) break;

      SetExclusion(ai, bi);
    }
  }

  CopyExclusionListToDevice();

}

///////////////////////////////////////////
// Copy functions
///////////////////////////////////////////

void NeighborList::CopyExclusionListToDevice(){
  unsigned int num_units = numberAllocatedUnits["ExclusionList"];
  if(num_units)
    cudaMemcpy( d_ExclusionList, h_ExclusionList, sizeof(NBTYPEv) * num_units, cudaMemcpyHostToDevice );
}
 
void NeighborList::CopyExclusionTypeToDevice(){
  unsigned int num_units = numberAllocatedUnits["ExclusionType"];
  if(num_units)
    cudaMemcpy( d_ExclusionType, h_ExclusionType, num_units * sizeof(unsigned), cudaMemcpyHostToDevice );
}

///////////////////////////////////////////////////
// Print information about the internal structure.
///////////////////////////////////////////////////

// 32 bit assumption on bit-pattern. The interaction matrix is layed out in a non-standard pattern.
void NeighborList::PrintExclusionList(){
  if(numberAllocatedUnits["ExclusionList"] == 0)
    throw RUMD_Error("NeighborList","PrintExclusionList","Exclusion list has not been allocated");

  cudaMemcpy( h_ExclusionList, d_ExclusionList, sizeof(NBTYPEv) * numBlocksParticles, cudaMemcpyDeviceToHost );
  
  for(unsigned i=0; i < numParticles; i++){
    cout << endl << endl << "Particle " << i << ": ";
    const unsigned int neighborListOffset = ( i / kp.threads.x ) * kp.num_blocks * kp.threads.x + ( i % kp.threads.x );

    for( unsigned j=0; j < numBlocks; j++ ){
      unsigned value = h_ExclusionList[neighborListOffset + j*kp.threads.x].x;
      
      if( value == 0 ){
	cout << "|0|";
      } // 2^32 - 1.
      else if(value == 4294967295){
	cout << "|1|";
      }
      else{ // Expand only non-trivial patterns.
	cout << "|";
	for(int z=kp.threads.x-1; z >= 0; z--){
	  unsigned m = 1 << z;
	  
	  if( ( value & m ) == 0 )
	    cout << "0";
	  else
	    cout << "1";
	}
	cout << "|";
      }
    }
  }
  cout << endl << endl;
} 


void NeighborList::PrintCellList() {
  if(numberAllocatedUnits["cell_list"] == 0)
    throw  RUMD_Error("NeighborList","PrintCellList","cell_list has not been allocated");

  unsigned int nb = numBlocks;
  int* h_cell_list = new int[nb*maxNumNbrBlocks];
  int EmptyCells = 0;
  
  CUDA_SAFE_CALL( cudaMemcpy( h_cell_list, cell_list, sizeof(int)*nb*maxNumNbrBlocks, cudaMemcpyDeviceToHost) );
  
  std::cout << "Cell-list:"  << std::endl;
  for (unsigned int i=0; i<nb; i++) {
    std::cout << i << ": ";
    for (unsigned j=0; j<maxNumNbrBlocks; j++) {
      if (h_cell_list[i*maxNumNbrBlocks + j]<0) {
	EmptyCells++;
	std::cout << ".";
      } else
	std::cout << h_cell_list[i*maxNumNbrBlocks + j] << " ";
    }
    std::cout << std::endl;
  }
  std::cout << "Non-empty cells per block:" << float(nb*nb - EmptyCells)/float(nb) << std::endl << std::endl;;

  delete[] h_cell_list;
}


void NeighborList::PrintBoundingboxes() { 
  if(numberAllocatedUnits["d_Rc_d_dR"] == 0)
    throw  RUMD_Error("NeighborList","PrintCellList","bounding box arrays have not been allocated");
 
  unsigned int nb = numBlocks;
  float4* h_Rc = new float4[nb]; 
  float4* h_dR = new float4[nb];  
  
  CUDA_SAFE_CALL( cudaMemcpy( h_Rc, d_Rc, sizeof(float4)*nb, cudaMemcpyDeviceToHost) );
  CUDA_SAFE_CALL( cudaMemcpy( h_dR, d_dR, sizeof(float4)*nb, cudaMemcpyDeviceToHost) );
  
  std::cout << "Boundingboxes:"  << std::endl;
  for (unsigned int i=0; i<nb; i++) {
    std::cout << h_Rc[i].x << " ";
    std::cout << h_Rc[i].y << " ";
    std::cout << h_Rc[i].z << "  ";
    std::cout << h_dR[i].x << " ";
    std::cout << h_dR[i].y << " ";
    std::cout << h_dR[i].z <<  std::endl;
  }
  std::cout << std::endl;
  
  delete[] h_Rc;
  delete[] h_dR;
}

///////////////////////////////////////////////
// Kernel Implementations
///////////////////////////////////////////////

// the following kernel is for when using LeesEdwardsSimulationBox
__global__ void update_deltaStrain(float* lastBoxShift, float2* dStrain, float* simBoxPointer, float skin, float max_cutoff) {
  float Lx = simBoxPointer[1];
  float Lx_inv = simBoxPointer[4];
  float Ly_inv = simBoxPointer[5];

  // change in boxShift
  float deltaStrain = simBoxPointer[7] - lastBoxShift[0];
  deltaStrain -= Lx * rintf( deltaStrain * Lx_inv );
  deltaStrain *= Ly_inv; // convert to strain
  
  float2 tmp = {0.f, 0.f};
  tmp.x = deltaStrain;
  float skin_adj = skin - deltaStrain* max_cutoff; 
  if (skin_adj < 0.) skin_adj = 0.;
  tmp.y = 0.25f*(skin_adj*skin_adj);
  dStrain[0] = tmp;
}

// When to update the neighborlist.
template <class S> __global__ void displacement_since_nb_build( unsigned numParticles, float4* r, float4* last_r, float4* image, 
								S* simBox, float* simBoxPointer, float skin2_over_4, int* updateNbList, float2* dStrain ){
  if( MyGP < numParticles ){
    float array[simulationBoxSize];
    simBox->loadMemory( array, simBoxPointer );
    
    // Distance moved since last update.
    float distanceMoved2 = (simBox->calculateDistanceMoved(r[MyGP], last_r[MyGP], array, dStrain)).w;

    // Need update?
    float s2_4 = skin2_over_4;

    if(dStrain != 0)
      s2_4 = dStrain[0].y;
    
    if( distanceMoved2 > s2_4 ){
      // The number of blocks MUST be the same as for calls to md_update_nblist!
      updateNbList[0] = NumBlocks;
      image[MyGP].w = 0.f;
    }
    else{
      // Consistency check performed when configuration IO. 
      image[MyGP].w = distanceMoved2;
    }
  }
}

// Calculate bounding box.
template <class S> __global__ void calc_boundingbox( unsigned int num_part, float4* r, float4* Rc, float4* dR, 
						     S* simBox, float* simBoxPointer, float maxCutPlusSkin, int* updateNbList ){
  if(updateNbList[0]){
    // Load the simulation box in local memory to avoid bank conflicts.
    float array[simulationBoxSize];
    simBox->loadMemory( array, simBoxPointer );
    
    extern __shared__ float4 s_r[];
    __shared__ float4 s_r_ref[1]; 
    
    float4 my_r;
    
    // Load coordinates
    if (MyGP < num_part) {
      my_r = r[MyGP];
    }

    // Perform minimum image convention with reference to particle 0 (which is always there)
    if (MyP==0) s_r_ref[0] = my_r;
    __syncthreads();
    
    if (MyGP < num_part) {
      s_r[MyP] = simBox->calculateDistance(my_r, s_r_ref[0], simBoxPointer);
    }
    __syncthreads();
        
    // Simple linear search. COULD be optimized ...
    if (MyP == 0) {
      float4 MinR = {0.0f, 0.0f, 0.0f, 0.0f};
      float4 MaxR = {0.0f, 0.0f, 0.0f, 0.0f};

      for (unsigned i=0; i<PPerBlock; i++) {
	if (MyGP+i < num_part) {   
	  if      (s_r[i].x < MinR.x) MinR.x = s_r[i].x;
	  else if (s_r[i].x > MaxR.x) MaxR.x = s_r[i].x;
	  if      (s_r[i].y < MinR.y) MinR.y = s_r[i].y;
	  else if (s_r[i].y > MaxR.y) MaxR.y = s_r[i].y;
	  if      (s_r[i].z < MinR.z) MinR.z = s_r[i].z;
	  else if (s_r[i].z > MaxR.z) MaxR.z = s_r[i].z;
	}
      }

      float4 temp;
      temp.x = (MaxR.x - MinR.x + maxCutPlusSkin)*0.5f;
      temp.y = (MaxR.y - MinR.y + maxCutPlusSkin)*0.5f;
      temp.z = (MaxR.z - MinR.z + maxCutPlusSkin)*0.5f;

      my_r.x =  s_r_ref[0].x + MinR.x + temp.x;
      my_r.y =  s_r_ref[0].y + MinR.y + temp.y;
      my_r.z =  s_r_ref[0].z + MinR.z + temp.z;

      Rc[MyB] = my_r; // Center of rectangle
      dR[MyB] = temp; // Distance from surfaces to center
    }
  }
}

template <class S> __global__ void calc_cell_list( unsigned int num_blocks, unsigned int maxNumNbrBlocks, float4* Rc, float4* dR, 
						   S* simBox, float* simBoxPointer, int* cell_list, int* updateNbList, unsigned int* num_nbr_blocks ){
  if(updateNbList[0]){
        // Load the simulation box in local memory to avoid bank conflicts.
    float simBoxPtr_local[simulationBoxSize];
    simBox->loadMemory( simBoxPtr_local, simBoxPointer );
        
    extern __shared__ float4 s_Rc[];
    float4* s_dR = (float4*) &s_Rc[blockDim.x];

    float4 my_Rc; float4 my_dR;

    // Read own coordinates
    unsigned int my_global_index = threadIdx.x + blockIdx.x*blockDim.x;
    if (my_global_index < num_blocks){
      my_Rc = Rc[my_global_index];
      my_dR = dR[my_global_index];
    }
    
    unsigned int Counter = 0;
    
    // Loop over all hyper-blocks
    for (unsigned int i=0; i<gridDim.x; i++){
      
      // Read coordinates of i'th hyper-block
      unsigned int my_read_index = threadIdx.x + i*blockDim.x;
      if (my_read_index < num_blocks) {
	s_Rc[threadIdx.x] = Rc[my_read_index];
	s_dR[threadIdx.x] = dR[my_read_index];
      }
      __syncthreads();

      // Check for overlaps
      if (my_global_index < num_blocks && Counter <= maxNumNbrBlocks){
	for (unsigned int j=0; j<blockDim.x; j++){
	  if (j + i*blockDim.x < num_blocks){
	    
	    float4 Dist = simBox->calculateDistance(my_Rc, s_Rc[j], simBoxPtr_local);
	    if ( abs(Dist.x)<(my_dR.x + s_dR[j].x) && abs(Dist.y)<(my_dR.y + s_dR[j].y)  && abs(Dist.z)<(my_dR.z + s_dR[j].z) ) {
	      // Found an overlap
	      
	      if(Counter == maxNumNbrBlocks) {
		Counter++; // increment so can detect overflow later
		break; // No room for this overlap; break out of loop for(j...) (shouldn't break out of for (i...) because of synchronization)
	      }
	      // OLD cell_list[ my_global_index*num_blocks + Counter] = i*blockDim.x + j;
	      cell_list[ my_global_index*maxNumNbrBlocks + Counter] = i*blockDim.x + j;// NEW
	      Counter++;
	      
	    } // if (abs....
	  } // if(j + ....
	} // for(j=0 ...
      } // if (my_global_index
    } // loop over hyper-blocks

    // Pad cell_list with '-1'
    if (my_global_index < num_blocks) {
    num_nbr_blocks[my_global_index] = Counter;
    // OLD for (; Counter<num_blocks; Counter++) {
    for (; Counter<maxNumNbrBlocks; Counter++) {
      // OLD cell_list[my_global_index*num_blocks + Counter] = -1;
      cell_list[my_global_index*maxNumNbrBlocks + Counter] = -1;
      }
    }
  }
}

template <class S> __global__ void md_update_nblist( unsigned int num_part, unsigned int max_num_nbr_blocks, float4* r, float4* im, float* params, 
						     S* simBox, float* simBoxPointer, NBTYPEv* NbList, NBTYPEv* ExclusionList, unsigned* exclusionType, 
						     int* cell_list, int* updateNbList, float4* last_r, float* lastBoxShift, unsigned numberOfTypes ){
  if(updateNbList[0]){
    const unsigned int offset = MyT * PPerBlock;
    // OLD const unsigned int CellOffset = MyB * NumBlocks * PPerBlock; 
    const unsigned int CellOffset = MyB * max_num_nbr_blocks * PPerBlock;

    // Exclusion list offset.
    const unsigned exclusionOffset = ( MyGP / PPerBlock ) * NumBlocks * PPerBlock + MyGP % PPerBlock;
    
    // Load the simulation box in local memory to avoid bank conflicts.
    float simBoxPtr_local[simulationBoxSize];
    simBox->loadMemory( simBoxPtr_local, simBoxPointer );
    
    // Read parameters into shared memory
    __shared__ float s_params[MaxNumTypes*MaxNumTypes]; 

    // Read exlusion types into shared memory.
    __shared__ unsigned localExclusionTypes[MaxNumTypes*MaxNumTypes]; 
    
    // Not secure!
    if (MyP<MaxNumTypes*MaxNumTypes && MyT==0){ 
      s_params[MyP] = params[MyP]*params[MyP];
    }

    if (MyP<numberOfTypes*numberOfTypes && MyT==0){ 
      localExclusionTypes[MyP] = exclusionType[MyP];
    }

    extern __shared__ float4 s_r[];

    float4 my_r;
    
    if (MyT==0) { 
      my_r = r[MyGP];
      s_r[MyP] = my_r;
    }
    
    __syncthreads();
    
    my_r = s_r[MyP];
    int MyType = __float_as_int(my_r.w);
    __syncthreads();
    
    float* my_params = &s_params[MaxNumTypes*__float_as_int(my_r.w)];

    // Run through cells in cell_list
    int cell = 0;
    
    unsigned int j;
    // OLD for (j=0; j<NumBlocks && cell>=0; j+=TPerPart) {
    for (j=0; j<max_num_nbr_blocks && cell>=0; j+=TPerPart) {
    
      // Read the cell-number of interacting cell
      cell = -1;
      // OLD if (j + MyT < NumBlocks) {
      if (j + MyT < max_num_nbr_blocks) {
        //OLD cell = cell_list[MyB*NumBlocks + j + MyT];
	cell = cell_list[MyB*max_num_nbr_blocks + j + MyT];
      }
    
      int Pread = cell*PPerBlock + MyP;

      // Exclusion list offset.
      unsigned eOffset = (Pread / PPerBlock) * PPerBlock;

      if (cell>=0 && cell<NumBlocks) {
	if (Pread<num_part) {
	  s_r[MyP + offset] = r[Pread]; 
	} else {
	  s_r[MyP + offset].w = -1.0f; // Mark particle as being 'dummy'
	}
      }    
    
      // Make sure all shared data loaded before using
      __syncthreads();  
    
      // Check all interactions in this cell
      NBTYPE NbListCell = 0;
      NBTYPE NbListAdd = 1;
      if (cell>=0 && cell<NumBlocks){
	if( MyGP < num_part ){     // If virtual just write 0 to NbList.
	  
	  // Loop over particles in cell
	  for (unsigned int i = 0; i < PPerBlock; i++, NbListAdd*=2){
	    float4 sr = s_r[i+offset];
	    
	    // Check for interaction (if other particle is not virtual and not self)
	    bool nbr_is_self = (cell == MyB && i == MyP);
	    if (sr.w>-0.5f && !nbr_is_self) {
	      float dist2 = (simBox->calculateDistance(my_r, sr, simBoxPtr_local)).w;
	      int OtherType = __float_as_int(sr.w);

	      if ( dist2 <= my_params[MyType] && !(localExclusionTypes[MyType*numberOfTypes + OtherType]) ){
		//NbListCell += (1<<i); 
		NbListCell += NbListAdd; 
	      }
	    }
	  }
	  
	  // Eliminate interactions.
	  if(ExclusionList)
	    NbListCell &= ~ExclusionList[exclusionOffset + eOffset].x;
	}
	
	// Write the nb-list for this cell to global memory
	// Note CellOffset involves maxNumNbrBlocks now
	NbList[CellOffset + (j+MyT)*PPerBlock + MyP].x = NbListCell;
      }
      
      // Make sure all threads are done with this cell before continuing
      __syncthreads();
    } 
    
    // NbList updated
    if (MyT==0) {
      last_r[MyGP] = my_r;
      if (MyP==0) {
        atomicSub(updateNbList,1);
        if(lastBoxShift != 0)
          lastBoxShift[0] = simBoxPtr_local[7];
      }
    }
  }
}

///////////////////////////////////////////////
// Consistency for sorting
///////////////////////////////////////////////

__global__ void UpdateMyExclusionListAfterSorting( unsigned int* revIndex, NBTYPEv* oldExclusionList, NBTYPEv* exclusionList, unsigned numParticles ){
  const unsigned int cellOffset = MyB * NumBlocks * PPerBlock; 

  if( MyGP < numParticles ){
    unsigned localParticle = revIndex[MyGP];
    unsigned newCellOffset = ( localParticle / PPerBlock ) * NumBlocks * PPerBlock + localParticle % PPerBlock;

    for (unsigned i=0; i < NumBlocks; i++){
      unsigned Index = cellOffset + i*PPerBlock + MyP;
      NBTYPE exclusionPattern = oldExclusionList[Index].x;
      
      int whichParticle = -1; unsigned numInteraction;
      
      while( exclusionPattern > 0 ){
	whichParticle += (numInteraction = __ffs(exclusionPattern));
	
	unsigned myParticle = i * PPerBlock + whichParticle;
	unsigned newParticle = revIndex[myParticle];
	
	// Flip the bit of newParticle to no interaction i.e. 0 to 1.
	unsigned newIndex = newCellOffset + (newParticle / PPerBlock) * PPerBlock;
	unsigned newPattern = ( 1 << ( newParticle % PPerBlock ) );
	exclusionList[newIndex].x |= newPattern; 
	
	exclusionPattern = (exclusionPattern >> numInteraction);   
      }
    }
  }
}
