/*
    Copyright (C) 2010  Thomas Schrøder

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    LICENSE.txt file for license details.
*/

#include "rumd/Sample.h"
#include "rumd/Integrator.h"
#include "rumd/IntegratorNVT.h"
#include "rumd/Potential.h"
#include "rumd/PairPotential.h"
#include "rumd/AnglePotential.h"
#include "rumd/DihedralPotential.h"
#include "rumd/ConstraintPotential.h"
#include "rumd/ConfigurationOutputManager.h"
#include "rumd/RUMD_Error.h"
#include "rumd/MoleculeData.h"
#include "rumd/Device.h"

#include <thrust/device_vector.h>
#include <thrust/sequence.h>
#include <thrust/sort.h>
#include <thrust/gather.h>

#include <algorithm>

/////////////////////////////////////////////
// CONSTRUCTION/DESTRUCTION 
/////////////////////////////////////////////

Sample::Sample( unsigned int pb, unsigned int tp, bool allow_tp_280_large=false ) 
  : particleData(),
    moleculeData(0),
    trajectoryDir("./TrajectoryFiles"),
    blockSize(0),
    simulationBox(0),
    own_sim_box(true),
    itg(0),
    verbose(true),
    sortingScheme(SORT_XY)
{
  cudaDeviceSetCacheConfig( cudaFuncCachePreferL1 );
  this->allow_tp_280_large = allow_tp_280_large;
  SetPB_TP(pb, tp);

  degreesOfFreedom = 0;

  for(unsigned int type = 0; type < MaxNumTypes; type++)
    GetParticleData()->SetMass(type, 1.0);

  CreateDefaultOutputManagers();
}

Sample::~Sample(){
  delete restartWriter;
  delete restartLogLin;
  
  delete trajectoryWriter;
  delete trajectoryLogLin;

  delete energiesLogLin;
  
  if(simulationBox && own_sim_box)
    delete simulationBox;
}

// Copy-constructor
Sample::Sample(const Sample& S) : particleData(),
				  trajectoryDir("./TrajectoryFiles"),
				  blockSize(0),
				  itg(0),
				  verbose(true) {
  this->allow_tp_280_large = S.allow_tp_280_large;
  SetPB_TP(S.GetParticlesPerBlock(), S.GetThreadsPerParticle());

  // num_sortingblocks_y and num_sortingblocks_x should be copied also.
  simulationBox = S.GetSimulationBox()->MakeCopy();
  own_sim_box = true;

  SetNumberOfParticles(S.GetNumberOfParticles());
  // we assume the device data is always the most up to date
  S.CopySimulationDataFromDevice();
  particleData.CopyHostData(*S.GetParticleDataConst()); // copies only on CPU
  CopySimulationDataToDevice();

  // may want to make it optional to do the following, or have a different
  // trajectoryDir
  CreateDefaultOutputManagers();
}

// here would like to be able to check a "DataStatus" flag to see if
// need to copy data from the device
Sample* Sample::Copy() {
  Sample* new_sample = new Sample( (*this) );
  return new_sample;
}

Sample& Sample::operator=(const Sample& S){
  if(this != &S){ 
    if(GetNumberOfParticles() != S.GetNumberOfParticles())
      SetNumberOfParticles(S.GetNumberOfParticles());

    // Note: the following only copies on the device.
    particleData = *S.GetParticleDataConst();
    
    if(!simulationBox) {
      simulationBox = S.GetSimulationBox()->MakeCopy();
      own_sim_box = true;
    }
    else
      simulationBox->CopyAnotherBox(S.GetSimulationBox());

    // Trigger neighborList update in any pair potential(s).
    std::vector<Potential*>::iterator potIter;
    for(potIter = potentialList.begin(); potIter != potentialList.end(); potIter++) {
      PairPotential* pair_pot = dynamic_cast<PairPotential*>(*potIter);
      if(pair_pot)
	pair_pot->ResetNeighborList();
    }
  }
  return *this; 
}

/////////////////////////////////////////////
// Initialization of Sample 
/////////////////////////////////////////////

void Sample::SetPB_TP( unsigned int pb, unsigned int tp ) {

  if( !( (pb != 0) && (pb <= 32) && ( (pb & (~pb + 1)) == pb ) ) )
    throw(RUMD_Error("Sample","SetPB_TP","pb is not a power of 2 or is larger than 32"));
  
  threads_per_particle = tp;
  particles_per_block = pb;
  
  unsigned int current_nParticles = GetNumberOfParticles();
  if(current_nParticles)
    SetNumberOfParticles(current_nParticles);
}

void Sample::CreateDefaultOutputManagers() {
  writers_initialized = false;
  
  restartWriter = new ConfigurationWriterReader();
  restartWriter->metaData.SetPrecision(9);
  restartWriter->metaData.Set("images", true);
  restartWriter->metaData.Set("velocities", true);
  restartWriter->metaData.Set("forces", true);
  restartWriter->metaData.Set("pot_energies", true);
  restartWriter->metaData.Set("virials", true);
  
  restartLogLin = new ConfigurationOutputManager(this, trajectoryDir, "restart", restartWriter);
  restartLogLin->SetCreatesDirectory(true);
  restartLogLin->SetVerbose(verbose);
  outputManagers.push_back(restartLogLin);

  trajectoryWriter = new ConfigurationWriterReader();
  trajectoryWriter->metaData.SetPrecision(4);
  trajectoryWriter->metaData.Set("images", true);
  trajectoryWriter->metaData.Set("velocities", false);
  trajectoryWriter->metaData.Set("forces", false);
  trajectoryWriter->metaData.Set("pot_energies", true);
  trajectoryWriter->metaData.Set("virials", true);
  trajectoryWriter->metaData.Set("logLin", true);  

  trajectoryLogLin = new ConfigurationOutputManager(this, trajectoryDir, "block", trajectoryWriter);
  trajectoryLogLin->SetDuplicateBlockEnds(true);
  trajectoryLogLin->SetVerbose(verbose);
  // make the default behaviour pure logarithmic saving with base 1:  
  trajectoryLogLin->SetLogLinParams(1,0);
  outputManagers.push_back(trajectoryLogLin);

  energiesLogLin = new EnergiesOutputManager(this, trajectoryDir, "energies");
  energiesLogLin->metaData.precision = 6;
  energiesLogLin->SetVerbose(verbose);
  outputManagers.push_back(energiesLogLin);
}

void Sample::SetSimulationBox(SimulationBox* simBox, bool own_this_box){ 
  if(own_sim_box) delete simulationBox;
  simulationBox = simBox;

  for(std::vector<Potential*>::iterator potIter = potentialList.begin(); potIter != potentialList.end(); potIter++)
    (*potIter)->SetSampleParameters(kPlan, &particleData, simulationBox);

  own_sim_box = own_this_box;
}

void Sample::SetNumberOfParticles(unsigned int set_num_part){

  // avoid problem with 280 cards/large sizes/ tp >1
  const std::string device_name = Device::GetDevice().GetDeviceName();
  if(set_num_part >= 8000 && (device_name == "Tesla C1060" || device_name == "GeForce GTX 280") && threads_per_particle > 1)
    if(allow_tp_280_large)
      std::cerr << "WARNING: Problems for large sizes on 280 cards with tp>1" << std::endl;
    else
      throw RUMD_Error("Sample","SetNumberOfParticle",std::string("Aborting due to problems with these cards for large systems and tp>1. Set tp=1 instead. Alternatively pass allow_tp_280_large=True as an argument for rumdSimulation (at your own risk). Device=") + device_name);

  particleData.SetNumberOfParticles(set_num_part, particles_per_block);
  // Print useful info.
  if(verbose) {
    std::cout << std::endl << "Setting number of particles and allocating on host and device:" << std::endl;
    std::cout << "Number of particles (N): " << std::setw( 9 ) << particleData.GetNumberOfParticles() << std::endl;
    std::cout << "Particles per block (pb): " << std::setw( 6 ) <<  particles_per_block << std::endl;
    std::cout << "Threads per particle (tp): " << std::setw( 4 )<<  threads_per_particle << std::endl;
    std::cout << "Number of blocks: " <<  std::setw( 14 ) << particleData.GetNumberOfBlocks() << std::endl;
    std::cout << "Virtual particles: " <<  std::setw( 15 ) << particleData.GetNumberOfVirtualParticles() << std::endl;
    std::cout << "Unused virtual particles: " << std::setw( 6 ) << (particleData.GetNumberOfVirtualParticles() - particleData.GetNumberOfParticles()) << std::endl;
    std::cout << "Threads per block: " << std::setw( 14 ) << (threads_per_particle*particles_per_block) << std::endl;
    std::cout << "Number of threads: " <<  std::setw( 15 ) << (threads_per_particle*particles_per_block*particleData.GetNumberOfBlocks()) << std::endl << std::endl;
  }

  degreesOfFreedom = 3*particleData.GetNumberOfParticles() - 3; 

  kPlan.grid.x = particleData.GetNumberOfBlocks();
  kPlan.threads.x = particles_per_block;
  kPlan.threads.y = threads_per_particle;
  kPlan.shared_size = particles_per_block * threads_per_particle * sizeof(float4);
  kPlan.num_blocks = particleData.GetNumberOfBlocks();
  kPlan.num_virt_part = particleData.GetNumberOfVirtualParticles();
  
  for(std::vector<Potential*>::iterator potIter = potentialList.begin(); potIter != potentialList.end(); potIter++)
    (*potIter)->SetSampleParameters(kPlan, &particleData, simulationBox);
  
  if(itg)
    itg->SetSampleData(this, &particleData);
}

void Sample::SetPotential(Potential* potential){ 
  potentialList.clear();
  AddPotential(potential);
}

void Sample::AddPotential(Potential* potential){
  potential->SetVerbose(verbose);

  if(GetNumberOfParticles())
    potential->SetSampleParameters(kPlan, &particleData, simulationBox);

  if(dynamic_cast<ConstraintPotential*>(potential)){
    potentialList.push_back(potential);
  }
  else{
    // Will induce reallocation, but this is a small list.
    potentialList.insert(potentialList.begin(), potential);
  }
  energiesLogLin->metaData.options[potential->GetID_String()] = false;
  energiesLogLin->metaData.fileStr[potential->GetID_String()] = potential->GetID_String();

}

// Set the current integrator.
void Sample::SetIntegrator(Integrator* newItg){
  itg = newItg;
  
  // SetNumberOfParticles has been called earlier and did not allocate.
  if(particleData.GetNumberOfParticles())
    itg->SetSampleData(this, &particleData);
}

///////////////////////////////////////////////////////////////
// Force and potential
///////////////////////////////////////////////////////////////

float Sample::CalcF(bool calc_stresses, bool report_time){
  if(!simulationBox)
    throw RUMD_Error("Sample","CalcF","No simulation box has been set");
  if(!potentialList.size())
    std::cerr << "Warning [Sample::CalcF]: No potential has been set" << std::endl;
  

  float elapsedTime = 0.f;
  for(unsigned int pdx=0; pdx < potentialList.size(); ++pdx)
    elapsedTime += potentialList[pdx]->CalcF(pdx==0, calc_stresses, report_time);
  if(itg)
    itg->CalculateAfterForce();


 cudaError err = cudaGetLastError();
 if(err != cudaSuccess) {
   if(err == cudaErrorLaunchOutOfResources ||
      err == cudaErrorInvalidConfiguration)
     throw( RUMD_cudaResourceError("Sample","CalcF",std::string("tp too high: ") + cudaGetErrorString(err)) );
   else   
     throw( RUMD_Error("Sample","CalcF",std::string("cuda error: ") + cudaGetErrorString(err) ) );
    }

 return elapsedTime;
}

double Sample::GetPotentialEnergy(bool copy) const {

  if(copy)
    particleData.CopyForFromDevice();

  unsigned int nParticles = GetNumberOfParticles();

  double potentialEnergy = 0.;
  for(unsigned int i=0; i < nParticles; i++)
    potentialEnergy += particleData.h_f[i].w;

  // loop over potentials
  for(unsigned int pdx=0; pdx < potentialList.size(); ++pdx) {
    if(!potentialList[pdx]->EnergyIncludedInParticleSum())
      potentialEnergy += potentialList[pdx]->GetPotentialEnergy();
  }

  return potentialEnergy;
}

double Sample::GetVirial(bool copy) const {
  if(copy)
    particleData.CopyVirFromDevice();
  double virial = 0.;
  unsigned int nParticles = GetNumberOfParticles();

  for(unsigned int i=0; i < nParticles; i++) { 
    virial += particleData.h_w[i].w;
  }
  return virial/6.0;
}

////////////////////////////////////////////////////////
// IO   
////////////////////////////////////////////////////////

void Sample::SetVerbose(bool vb) {
  verbose = vb;
  std::vector<LogLinOutputManager*>::iterator llIter;
  for(llIter = outputManagers.begin(); llIter != outputManagers.end(); llIter++)
    (*llIter)->SetVerbose(verbose);

  for(unsigned int pdx=0; pdx < potentialList.size(); ++pdx)
    potentialList[pdx]->SetVerbose(verbose);

}

void Sample::SetOutputDirectory(const std::string& trajectoryDir) {
  this->trajectoryDir = trajectoryDir;
  std::vector<LogLinOutputManager*>::iterator llIter;
  for(llIter = outputManagers.begin(); llIter != outputManagers.end(); llIter++)
    (*llIter)->SetOutputDirectory(trajectoryDir);
  
  writers_initialized = false;
}

void Sample::SetOutputBlockSize(unsigned long int blockSize) {
  if(!blockSize) throw RUMD_Error("Sample","SetOutputBlockSize","blockSize must be greater than zero");
  this->blockSize = blockSize;
  
  std::vector<LogLinOutputManager*>::iterator llIter;
  for(llIter = outputManagers.begin(); llIter != outputManagers.end(); llIter++)
    (*llIter)->SetBlockSize(blockSize);

  restartLogLin->SetLogLinParams(blockSize,0);

  writers_initialized = false;
}

void Sample::SetWriteEnergies(bool writeEnergies) {
  std::vector<LogLinOutputManager*>::iterator ellIt = std::find(outputManagers.begin(), outputManagers.end(), energiesLogLin);
  bool energiesOnList = ( ellIt != outputManagers.end() );

  if(writeEnergies) {
    energiesLogLin->SetVerbose(verbose);
    energiesLogLin->SetBlockSize(blockSize);
  }

  if(writeEnergies && !energiesOnList)
    outputManagers.push_back(energiesLogLin);
  else if(!writeEnergies && energiesOnList)
    outputManagers.erase(ellIt);
}

void Sample::SetWriteTrajectory(bool writeTrajectory) {
  std::vector<LogLinOutputManager*>::iterator ellIt = std::find(outputManagers.begin(), outputManagers.end(), trajectoryLogLin);
  bool trajOnList = ( ellIt != outputManagers.end() );

  if(writeTrajectory) {
    trajectoryLogLin->SetVerbose(verbose);
    trajectoryLogLin->SetBlockSize(blockSize);
  }
  
  if(writeTrajectory && !trajOnList)
    outputManagers.push_back(trajectoryLogLin);
  else if(!writeTrajectory && trajOnList)
    outputManagers.erase(ellIt);
}

void Sample::SetWriteRestart(bool writeRestart) {
  std::vector<LogLinOutputManager*>::iterator ellIt = std::find(outputManagers.begin(), outputManagers.end(), restartLogLin);
  bool restartOnList = ( ellIt != outputManagers.end() );

  if(writeRestart) {
    restartLogLin->SetVerbose(verbose);
    restartLogLin->SetBlockSize(blockSize);
  }
  
  if(writeRestart && !restartOnList)
    outputManagers.push_back(restartLogLin);
  else if(!writeRestart && restartOnList)
    outputManagers.erase(ellIt);
}

void Sample::AddOutputManager(LogLinOutputManager* om) {
  om->SetVerbose(verbose);
  om->SetBlockSize(blockSize);
  outputManagers.push_back(om);
}

void Sample::SetTrajectoryLinearWriteInterval(unsigned long int writeInterval) { 
  trajectoryLogLin->SetLogLinParams(writeInterval, writeInterval);
  trajectoryLogLin->SetDuplicateBlockEnds(false);
  writers_initialized = false;
}

void Sample::SetTrajectoryLogLinParams(unsigned int base, unsigned int maxInterval) {
  trajectoryLogLin->SetLogLinParams(base, maxInterval);
  trajectoryLogLin->SetDuplicateBlockEnds(true);
  writers_initialized = false;
}

void Sample::SetEnergyLinearWriteInterval(unsigned int writeInterval) {
  energiesLogLin->SetLogLinParams(writeInterval, writeInterval);
  energiesLogLin->metaData.writeInterval = writeInterval;
  energiesLogLin->metaData.has_logLin = false;
  energiesLogLin->SetDuplicateBlockEnds(false); 
  writers_initialized = false;
}

void Sample::SetEnergyLogLinParams(unsigned int base, unsigned int maxInterval) {
  energiesLogLin->SetLogLinParams(base, maxInterval);
  energiesLogLin->metaData.writeInterval = 0.;
  energiesLogLin->metaData.has_logLin = true;
  energiesLogLin->SetDuplicateBlockEnds(true); 
  writers_initialized = false;
}

void Sample::InitializeOutputManagers(unsigned long int timeStepIndex) {
  // iterate over vector calling Initialize on each output manager 
  std::vector<LogLinOutputManager*>::iterator fwIter;
  for(fwIter = outputManagers.begin(); fwIter != outputManagers.end(); fwIter++)
    (*fwIter)->Initialize(timeStepIndex);
  writers_initialized = true;
}

void Sample::NotifyOutputManagers(unsigned long int timeStepIndex) {
  if(!writers_initialized)
    InitializeOutputManagers(timeStepIndex);
  
  // iterate over vector calling Update 
  std::vector<LogLinOutputManager*>::iterator fwIter;
  for(fwIter = outputManagers.begin(); fwIter != outputManagers.end(); fwIter++)
    (*fwIter)->Update(timeStepIndex);

}

void Sample::TerminateOutputManagers() {
  if(!writers_initialized)
    std::cerr << "Warning: TerminateOutputManagers called although they have not been initialized" << std::endl;
  
  // iterate over vector calling Update 
  std::vector<LogLinOutputManager*>::iterator llIter;
  for(llIter = outputManagers.begin(); llIter != outputManagers.end(); llIter++)
    (*llIter)->Terminate();

}

void Sample::ReadRestartConf( int restartBlock, unsigned int numDigitsBlkIdx ) {
  // following works only on CUDA 2.2 or higher
  std::ostringstream restartFilePath;
  restartFilePath << trajectoryDir << "/restart" << std::setfill('0') 
		  << std::setw(numDigitsBlkIdx) << std::setprecision(numDigitsBlkIdx)  
		  << restartBlock << ".xyz.gz";
  
  ReadConf(restartFilePath.str(), true);
}

void Sample::ReadConf( const std::string& filename, bool init_itg ) {
  ConfigurationWriterReader reader;
  reader.Read(this, filename);
  if(itg && init_itg && reader.metaData.IntegratorFound()) {
    bool initialize = true;
    if(reader.metaData.GetIO_Format() == 1) {
       IntegratorNVT* testItg = 0;
       testItg = dynamic_cast<IntegratorNVT*>(itg);
       if(testItg)
	 testItg->SetThermostatState(reader.metaData.GetThermostatPs());
       else
	 initialize = false;
    }
    else
      itg->InitializeFromInfoString(reader.metaData.GetIntegratorInfoStr());

    if(initialize)
      std::cerr << "[Info] (Sample::ReadConf): Initializing integrator from parameters in file " << filename << std::endl;
  } // if(itg ...)
}

//////////////////////////////////////////////
// Simulation modification. !! Adjust for topology file. !!
//////////////////////////////////////////////

// Calculate the CM of a molecule. The first particle defines the molecule.
float4 Sample::CalculateCenterOfMass( float4* position, float4* velocity, int moleculeIndex ){
  RectangularSimulationBox* testRSB = dynamic_cast<RectangularSimulationBox*>(simulationBox);
  
  float moleculeMass = 1.f / velocity[moleculeIndex].w;
  float4 firstParticle = position[moleculeIndex];
  float4 CM = { moleculeMass * firstParticle.x, moleculeMass * firstParticle.y, moleculeMass * firstParticle.z, 0 }; 
  
  for(int i = moleculeIndex+1; i < moleculeIndex + 2; i++){ 
    float particleMass = 1.f / velocity[i].w;
    
    float4 difference = testRSB->calculateDistance( position[i], firstParticle, testRSB->GetHostPointer() );
    float3 nextParticle = { firstParticle.x + difference.x, firstParticle.y + difference.y, firstParticle.z + difference.z };
    
    CM.x += particleMass * nextParticle.x;
    CM.y += particleMass * nextParticle.y;
    CM.z += particleMass * nextParticle.z;
    moleculeMass += particleMass; 
  }
  
  CM.x /= moleculeMass;
  CM.y /= moleculeMass;
  CM.z /= moleculeMass;
  CM.w = ( 1.0 / moleculeMass );
  
  return CM;
}

// Simple particle scaling. Adjust for topology file.
void Sample::IsotropicScaleSystem( float Rscal ){

  particleData.CopyPosFromDevice();

  for(unsigned int i=0; i < particleData.GetNumberOfParticles(); i++){
    particleData.h_r[i].x *= Rscal;
    particleData.h_r[i].y *= Rscal;
    particleData.h_r[i].z *= Rscal;
  }
  
  simulationBox->ScaleBox( Rscal );  
  particleData.CopyPosToDevice();

}

void Sample::AnisotropicScaleSystem( float Rscal, int dir ){
  particleData.CopyPosFromDevice();

  for(unsigned int i=0; i < particleData.GetNumberOfParticles(); i++){
    if ( dir == 0 )
      particleData.h_r[i].x *= Rscal;
    else if ( dir == 1 )
      particleData.h_r[i].y *= Rscal;
    else if ( dir == 2 )
      particleData.h_r[i].z *= Rscal;
  }
  
  simulationBox->ScaleBoxDirection( Rscal, dir );  
  particleData.CopyPosToDevice();
}

//////////////////////////////////////////////
// Copy Function
//////////////////////////////////////////////

void Sample::CopySimulationDataToDevice() const {
  particleData.CopyConfToDevice();
}

void Sample::CopySimulationDataFromDevice() const{
  particleData.CopyConfFromDevice();
}

//////////////////////////////////////////////
// Sorting of particles on device. Experimental [TBS]
//////////////////////////////////////////////

__global__  void calc_sort_key( unsigned int num_part, float4* r, float *L, unsigned int Nx,  unsigned int Ny,  unsigned int Nz,  float* key){
  
  if( MyGP < num_part ){
    float4 My_r = r[MyGP];
   
    // Calculate key to be sorted after
    float iy = floor((My_r.y/L[2]/(1.+1.e-6)+0.5f)*Ny);
    float iz = floor((My_r.z/L[3]/(1.+1.e-6)+0.5f)*Nz);
    
    key[MyGP] = (iy + iz*Ny)*L[1] + My_r.x;
  }
}


void Sample::SortParticles(){
  // Do spatial sorting of particles
  unsigned int numParticles = particleData.GetNumberOfParticles();

  RectangularSimulationBox* testRSB = dynamic_cast<RectangularSimulationBox*>(simulationBox);
  if(!testRSB) 
    throw RUMD_Error("NeighborList", "SortParticles", "Can only do spatial sorting with RectangularSimulationBox");

  float4 simBox = testRSB->GetSimulationBox();

  unsigned int Ny = 1, Nz = 1;
  float Lx = simBox.x, Ly = simBox.y, Lz = simBox.z, V = simBox.w;

  if(sortingScheme == SORT_XY) {
    float alpha = pow(GetNumberOfParticles()/particles_per_block/Lx/Ly,0.5);
    Ny = (unsigned int) ( ceil(alpha*Ly) -1);
  }
  else if (sortingScheme == SORT_XYZ) {
    float alpha = pow(GetNumberOfParticles()/particles_per_block/V,1./3);
    Ny = (unsigned int) ( ceil(alpha*Ly) - 1);
    Nz = (unsigned int) ( ceil(alpha*Lz) - 1);
  }

  // wrap raw pointers with thrust device_ptrs 
  thrust::device_ptr<float4> thrust_d_r(particleData.d_r);
  thrust::device_ptr<float4> thrust_d_im(particleData.d_im);
  thrust::device_ptr<float4> thrust_d_v(particleData.d_v);
  thrust::device_ptr<float4> thrust_d_f(particleData.d_f);
  thrust::device_ptr<float4> thrust_d_w(particleData.d_w);
  thrust::device_ptr<float4> thrust_d_misc(particleData.d_misc);
  thrust::device_ptr<float4> thrust_d_temp(particleData.d_temp);
  thrust::device_ptr<unsigned int> thrust_d_unsorted_index(particleData.d_unsorted_index);
  thrust::device_ptr<unsigned int> thrust_d_temp_uint(particleData.d_temp_uint);

  // generate index vector to be used as value in sorting - used for updating data structures after sorting 
  thrust::device_vector<unsigned int> thrust_new_index(numParticles);
  thrust::sequence(thrust_new_index.begin(), thrust_new_index.end());

  // generate key vector 
  thrust::device_vector<float> thrust_key(numParticles);
  float* raw_ptr = raw_pointer_cast(&thrust_key[0]);
  
  int block_size = 256;
  int grid_size = (numParticles + block_size - 1) / block_size;
  
  // Fill key vector.
  calc_sort_key<<< grid_size, block_size >>>(numParticles, particleData.d_r, testRSB->GetDevicePointer(), 1, Ny, Nz, raw_ptr);

  // Sort key and values.
  thrust::sort_by_key(thrust_key.begin(), thrust_key.end(), thrust_new_index.begin());

  // Move all particle data to be consistent with new sorted order (Gather + Swap instead?)
  thrust::copy(thrust_d_r, thrust_d_r + numParticles, thrust_d_temp);
  thrust::gather(thrust_new_index.begin(), thrust_new_index.end(), thrust_d_temp, thrust_d_r);

  thrust::copy(thrust_d_im, thrust_d_im + numParticles, thrust_d_temp);
  thrust::gather(thrust_new_index.begin(), thrust_new_index.end(), thrust_d_temp, thrust_d_im);
  
  thrust::copy(thrust_d_v, thrust_d_v + numParticles, thrust_d_temp);
  thrust::gather(thrust_new_index.begin(), thrust_new_index.end(), thrust_d_temp, thrust_d_v);

  thrust::copy(thrust_d_f, thrust_d_f + numParticles, thrust_d_temp);
  thrust::gather(thrust_new_index.begin(), thrust_new_index.end(), thrust_d_temp, thrust_d_f);

  thrust::copy(thrust_d_w, thrust_d_w + numParticles, thrust_d_temp);
  thrust::gather(thrust_new_index.begin(), thrust_new_index.end(), thrust_d_temp, thrust_d_w);

  thrust::copy(thrust_d_misc, thrust_d_misc + numParticles, thrust_d_temp);
  thrust::gather(thrust_new_index.begin(), thrust_new_index.end(), thrust_d_temp, thrust_d_misc);
  
  thrust::copy(thrust_d_unsorted_index, thrust_d_unsorted_index + numParticles, thrust_d_temp_uint);
  thrust::gather(thrust_new_index.begin(), thrust_new_index.end(), thrust_d_temp_uint, thrust_d_unsorted_index);

  // Update data structure in MoleculeData is present and NeighborLists
  // (via potentials)
  unsigned int* d_raw_ptr_new_index = raw_pointer_cast(&thrust_new_index[0]);
  if(moleculeData)
    moleculeData->UpdateAfterSorting(d_raw_ptr_new_index);

  // Trigger update in any potential after sorting the particles.
  std::vector<Potential*>::iterator potIter;
  for(potIter = potentialList.begin(); potIter != potentialList.end(); potIter++)
    (*potIter)->UpdateAfterSorting(d_raw_ptr_new_index);
}

